"""Tests for the `Port` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import pytest
from marmot import Environment

from wisdem.orbit.core import Port, Cargo
from wisdem.orbit.core.exceptions import ItemNotFound


class SampleItem(Cargo):
    def __init__(self):
        pass


def test_port_creation():

    env = Environment()
    port = Port(env)
    item = SampleItem()

    port.put(item)
    port.put(item)

    items = [item for item in port.items if item.type == "SampleItem"]
    assert len(items) == 2


def test_get_item():

    env = Environment()
    port = Port(env)
    item = SampleItem()

    port.put(item)
    port.put(item)

    returned = port.get_item("SampleItem")
    assert returned == item
    assert len(port.items) == 1

    port.put({"type": "Not type Cargo"})
    with pytest.raises(ItemNotFound):
        _ = port.get_item("WrongItem")

    _ = port.get_item("SampleItem")
    with pytest.raises(ItemNotFound):
        _ = port.get_item("SampleItem")



import math
import traceback

import pandas as pd
from wisdem.landbosse.model.CostModule import CostModule


class DevelopmentCost(CostModule):
    """
    **DevelopmentCost.py

    -Created by Parangat Bhaskar on June 30, 2019

    Creating a simple DevelopmentCost module for now. This module reads in a user input from the detailed input Excel
    file and outputs this in the detailed output excel file.

    """

    def __init__(self, input_dict, output_dict, project_name):
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

    def calculate_costs(self):
        """
        Sets the total cost for development.

        If there is a key of 'development_labor_cost_usd' in the input
        dictionary, then that is used as the development cost and a
        dataframe is created that holds that labor cost.

        If the key is not present in the dictionary, the development
        cost is retrieved from the project data.

        Returns
        ------------------------------------
        total_development_cost : pd.DataFrame
            data frame with total development cost by type of cost (e.g., Labor)
        """
        if "development_labor_cost_usd" in self.input_dict:
            total_development_cost = pd.DataFrame(
                [
                    {"Type of cost": "Equipment rental", "Cost USD": 0, "Phase of construction": "Development"},
                    {
                        "Type of cost": "Labor",
                        "Cost USD": self.input_dict["development_labor_cost_usd"],
                        "Phase of construction": "Development",
                    },
                    {"Type of cost": "Materials", "Cost USD": 0, "Phase of construction": "Development"},
                    {"Type of cost": "Mobilization", "Cost USD": 0, "Phase of construction": "Development"},
                    {"Type of cost": "Other", "Cost USD": 0, "Phase of construction": "Development"},
                ]
            )
        else:
            total_development_cost = self.input_dict["development_df"]

        self.output_dict["total_development_cost"] = total_development_cost

        return total_development_cost

    def outputs_for_detailed_tab(self):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """

        result = []
        module = type(self).__name__
        for _, row in self.output_dict["total_development_cost"].iterrows():
            dashed_row = "{} - {} - {}".format(
                row["Type of cost"], row["Phase of construction"], math.ceil(row["Cost USD"])
            )
            result.append(
                {
                    "unit": "",
                    "type": "dataframe",
                    "variable_df_key_col_name": "Type of Cost - Phase of Construction - Cost in USD",
                    "value": dashed_row,
                    "last_number": row[2],
                }
            )

        for _dict in result:
            _dict["project_id_with_serial"] = self.project_name
            _dict["module"] = module

        self.output_dict["development_cost_csv"] = result
        return result

    def run_module(self):
        """
        Runs the DevelopmentCost module and populates the IO dictionaries with calculated values.

        """

        try:
            self.calculate_costs()
            self.outputs_for_detailed_tab()
            self.output_dict["development_module_type_operation"] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict["total_development_cost"], project_id=self.project_name, total_or_turbine=True
            )
            return 0, 0  # module ran successfully

        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} DevelopmentCost")
            return 1, error  # module did not run successfully



# playing around with seabed contact scenarios

# import pytest
# from numpy.testing import assert_allclose

import numpy as np
import moorpy as mp
import matplotlib.pyplot as plt

ms = mp.System(depth=200)

ms.addPoint(1, [-100, 0, -50])
ms.addPoint(1, [100, 0, 0])
ms.addPoint(0, [0, 0, -90])

ms.addLineType("chain", 0.1, 50.0, 1e12)

ms.addLine(120, "chain")
ms.addLine(200, "chain")

ms.pointList[2].attachLine(1, 0)
ms.pointList[2].attachLine(2, 0)
ms.pointList[0].attachLine(1, 1)
ms.pointList[1].attachLine(2, 1)

ms.initialize()

ms.solveEquilibrium3()
fig, ax = ms.plot2d(color="r")

for depth in [100, 110, 120, 130, 140]:
    ms.depth = depth
    ms.solveEquilibrium3()
    ms.plot2d(color="k", ax=ax)

plt.show()



import os
import unittest

import openmdao.api as om
# from wisdem.rotorse.rotor_cost import StandaloneRotorCost, initialize_omdao_prob
# from wisdem.glue_code.gc_LoadInputs import WindTurbineOntologyPython
# from wisdem.glue_code.gc_PoseOptimization import PoseOptimization
from wisdem import run_wisdem




class TestRC(unittest.TestCase):
    def testBladeJointSizerBAR_USC(self):

        wisdem_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
        example_dir = os.path.join(wisdem_dir, "examples", "03_blade")  # get path example 03_blade
        fname_modeling_options = os.path.join(example_dir, "modeling_options.yaml")
        fname_wt_input = os.path.join(example_dir, "BAR_USC.yaml")
        fname_analysis_options = os.path.join(example_dir, "analysis_options_no_opt.yaml")
        accuracy = 0
        wt_opt, _, _ = run_wisdem(fname_wt_input, fname_modeling_options,
                                                                fname_analysis_options)

        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.blade_mass"][0], 51002.33629166093, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.L_transition_joint"][0], -0.004017126893942291, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.t_reinf_ratio_joint"][0], 0.9899571827651443, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.w_reinf_ratio_joint"][0], 1.2087641190997038, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.n_joint_bolt"][0], 38.05368523091661, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.joint_mass"][0], 430.94139294963134, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.joint_material_cost"][0], 1857.5411987710845, places=accuracy)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestRC))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



"""`InstallPhase` base class."""

__author__ = ["Jake Nunemaker", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = ["Jake Nunemaker", "Rob Hammond"]
__email__ = ["jake.nunemaker@nrel.gov", "robert.hammond@nrel.gov"]


from abc import abstractmethod
from itertools import groupby

import numpy as np
import simpy
import pandas as pd

from wisdem.orbit.core import Port, Vessel, Environment
from wisdem.orbit.phases import BasePhase
from wisdem.orbit.core.defaults import common_costs


class InstallPhase(BasePhase):
    """BasePhase subclass for install modules."""

    def __init__(self, weather, **kwargs):
        """
        Creates an instance of `InstallPhase`.

        Parameters
        ----------
        weather : pd.DataFrame | np.ndarray
            Weather profile at site.
        """

        self.extract_phase_kwargs(**kwargs)
        self.initialize_environment(weather, **kwargs)
        self.availability = kwargs.get("availability", None)

    def initialize_environment(self, weather, **kwargs):
        """
        Initializes a `marmot.Environment` at `self.env`.

        Parameters
        ----------
        weather : np.ndarray
            Weather profile at site.
        """

        if isinstance(weather, pd.DataFrame):
            weather = weather.to_records()

        env_name = kwargs.get("env_name", "Environment")
        self.env = Environment(name=env_name, state=weather, **kwargs)

    def initialize_vessel(self, name, specs):
        """"""

        avail = getattr(self, "availability")
        if avail is None:
            return Vessel(name, specs)

        elif isinstance(avail, dict):
            default = avail.get("default", 1)
            return Vessel(name, specs, avail.get(name, default))

        else:
            return Vessel(name, specs, avail)

    @abstractmethod
    def setup_simulation(self):
        """
        Sets up the required simulation infrastructure

        Generally, this creates the port, initializes the items to be
        installed, and initializes the vessel(s) used for the installation.
        """

        pass

    def initialize_port(self):
        """
        Initializes a Port object with N number of cranes.
        """

        self.port = Port(self.env)

        try:
            cranes = self.config["port"]["num_cranes"]
            self.port.crane = simpy.Resource(self.env, cranes)

        except KeyError:
            self.port.crane = simpy.Resource(self.env, 1)

    def run(self, until=None):
        """
        Runs the simulation on self.env.

        Parameters
        ----------
        until : int, optional
            Number of steps to run.
        """

        self.env._submit_log({"message": "SIMULATION START"}, "DEBUG")
        self.env.run(until=until)
        self.append_phase_info()
        self.env._submit_log({"message": "SIMULATION END"}, "DEBUG")

    def append_phase_info(self):
        """Appends phase information to all logs in `self.env.logs`."""

        for l in self.env.logs:
            l["phase"] = self.phase

    @property
    def port_costs(self):
        """Cost of port rental."""

        port = getattr(self, "port", None)

        if port is None:
            return 0

        else:
            key = "port_cost_per_month"
            port_config = self.config.get("port", {})
            rate = port_config.get("monthly_rate", common_costs[key])

            months = self.total_phase_time / (8760 / 12)
            return months * rate

    @property
    def installation_capex(self):
        """Returns sum of all installation costs in `self.env.actions`."""

        return np.nansum([a["cost"] for a in self.env.actions]) + self.port_costs

    @property
    def total_phase_time(self):
        """Returns total phase time in hours."""

        return max([a["time"] for a in self.env.actions])

    @property
    @abstractmethod
    def detailed_output(self):
        """Returns detailed phase information."""

        pass

    @property
    def agent_efficiencies(self):
        """
        Returns a summary of agent operational efficiencies.
        """

        efficiencies = {}

        s = sorted(self.env.actions, key=lambda x: (x["agent"], x["action"]))
        grouped = {
            k: sum([i["duration"] for i in list(v)]) for k, v in groupby(s, key=lambda x: (x["agent"], x["action"]))
        }
        agents = list(set([k[0] for k in grouped.keys()]))
        for agent in agents:
            total = sum([v for k, v in grouped.items() if k[0] == agent])

            try:
                delay = grouped[(agent, "Delay")]
                e = (total - delay) / total

            except KeyError:
                delay = 0.0
                e = 1.0

            except ZeroDivisionError:
                e = 1.0

            if not 0.0 <= e <= 1.0:
                raise ValueError(f"Invalid efficiency for agent '{agent}'")

            name = str(agent).replace(" ", "_")
            efficiencies[f"{name}_operational_efficiency"] = e

        return efficiencies

    @staticmethod
    def get_max_cargo_mass_utilzations(vessels):
        """
        Returns a summary of cargo mass efficiencies for list of input `vessels`.

        Parameters
        ----------
        vessels : list
            List of vessels to calculate efficiencies for.
        """

        outputs = {}

        for vessel in vessels:
            name = vessel.name.replace(" ", "_")
            storage = getattr(vessel, "storage", None)
            if storage is None:
                print("Vessel does not have storage capacity.")
                continue

            outputs[f"{name}_cargo_mass_utilization"] = vessel.max_cargo_mass_utilization

        return outputs

    @staticmethod
    def get_max_deck_space_utilzations(vessels):
        """
        Returns a summary of deck space efficiencies for list of input `vessels`.

        Parameters
        ----------
        vessels : list
            List of vessels to calculate efficiencies for.
        """

        outputs = {}

        for vessel in vessels:
            name = vessel.name.replace(" ", "_")
            storage = getattr(vessel, "storage", None)
            if storage is None:
                print("Vessel does not have storage capacity.")
                continue

            outputs[f"{name}_deck_space_utilization"] = vessel.max_deck_space_utilization

        return outputs



import textwrap

import validation

mywidth = 70
myindent = " " * 4
wrapper = textwrap.TextWrapper(initial_indent=myindent, subsequent_indent=myindent, width=mywidth)
rsthdr = ["*", "#", "=", "-", "^", "~", "%"]


def get_type_string(indict):
    outstr = ""
    if indict["type"] == "number":
        outstr = "Float"
        if "unit" in indict.keys() and indict["unit"].lower() != "none":
            outstr += ", " + indict["unit"]
        elif "units" in indict.keys() and indict["units"].lower() != "none":
            outstr += ", " + indict["units"]

    elif indict["type"] == "integer":
        outstr = "Integer"

    elif indict["type"] == "boolean":
        outstr = "Boolean"

    elif indict["type"] == "string":
        if "enum" in indict.keys():
            outstr = "String from, " + str(indict["enum"])
        else:
            outstr = "String"

    elif indict["type"] == "array":
        outstr = "Array of "
        if indict["items"]["type"] == "number":
            outstr += "Floats"
            if "unit" in indict["items"].keys() and indict["items"]["unit"].lower() != "none":
                outstr += ", " + indict["items"]["unit"]
            elif "units" in indict["items"].keys() and indict["items"]["units"].lower() != "none":
                outstr += ", " + indict["items"]["units"]

        elif indict["items"]["type"] == "integer":
            outstr += "Integers"
        elif indict["items"]["type"] == "string":
            outstr += "Strings"
        elif indict["items"]["type"] == "boolean":
            outstr += "Booleans"

    return outstr


def get_description_string(indict):
    outstr = ""
    if "description" in indict.keys():
        outstr += wrapper.fill(indict["description"])

    if "default" in indict.keys():
        outstr += "\n\n" + myindent + "*Default* = " + str(indict["default"])

    if "minimum" in indict.keys():
        outstr += "\n\n" + myindent + "*Minimum* = " + str(indict["minimum"])
    elif indict["type"] == "array" and "minimum" in indict["items"].keys():
        outstr += "\n\n" + myindent + "*Minimum* = " + str(indict["items"]["minimum"])

    if "maximum" in indict.keys():
        outstr += myindent + "*Maximum* = " + str(indict["maximum"]) + "\n"
    elif indict["type"] == "array" and "maximum" in indict["items"].keys():
        outstr += "\n\n" + myindent + "*Maximum* = " + str(indict["items"]["maximum"])

    outstr += "\n"
    return outstr


class Schema2RST(object):
    def __init__(self, fname):
        self.fname = fname
        self.fout = fname.replace(".yaml", ".rst")
        self.yaml = validation.load_yaml(fname)
        self.f = None

    def write_rst(self):
        self.f = open(self.fout, "w")
        self.write_header()
        self.write_loop(self.yaml["properties"], 0, self.fname.replace("yaml", ""))
        self.f.close()

    def write_header(self):
        self.f.write("*" * 30 + "\n")
        self.f.write(self.fname + "\n")
        self.f.write("*" * 30 + "\n")
        if "description" in self.yaml.keys():
            self.f.write(self.yaml["description"] + "\n")

    def write_loop(self, rv, idepth, name, desc=None):
        self.f.write("\n")
        self.f.write("\n")
        self.f.write(name + "\n")
        print(idepth)
        if idepth > 0:
            self.f.write(rsthdr[idepth - 1] * 40 + "\n\n")
        self.f.write("\n")
        if not desc is None:
            self.f.write(desc + "\n\n")
        for k in rv.keys():
            print(k)
            if "type" in rv[k]:
                if rv[k]["type"] == "object" and "properties" in rv[k].keys():
                    k_desc = None if not "description" in rv[k] else rv[k]["description"]
                    self.write_loop(rv[k]["properties"], idepth + 1, k, k_desc)

                elif rv[k]["type"].lower() in ["number", "integer", "string", "boolean"]:
                    self.f.write(":code:`" + k + "` : " + get_type_string(rv[k]) + "\n")
                    self.f.write(get_description_string(rv[k]) + "\n")

                elif (
                    rv[k]["type"].lower() == "array"
                    and "type" in rv[k]["items"]
                    and rv[k]["items"]["type"] == "object"
                    and "properties" in rv[k]["items"].keys()
                ):
                    k_desc = None if not "description" in rv[k]["items"] else rv[k]["items"]["description"]
                    self.write_loop(rv[k]["items"]["properties"], idepth + 1, k, k_desc)

                elif rv[k]["type"].lower() == "array":
                    # rv[k]['items']['type'] in ['number','integer','string','boolean']):
                    self.f.write(":code:`" + k + "` : " + get_type_string(rv[k]) + "\n")
                    self.f.write(get_description_string(rv[k]) + "\n")


if __name__ == "__main__":
    for ifile in ["geometry_schema.yaml", "modeling_schema.yaml", "analysis_schema.yaml"]:
        myobj = Schema2RST(ifile)
        myobj.write_rst()
